package at.ac.uibk.gitsearch.service;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import javax.annotation.PostConstruct;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.MediaType;

import org.codeability.sharing.plugins.api.SharingPluginConfig;
import org.codeability.sharing.plugins.api.SharingPluginConfig.Action;
import org.glassfish.jersey.client.ClientConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.fasterxml.jackson.jaxrs.json.JacksonJsonProvider;

import at.ac.uibk.gitsearch.config.ApplicationProperties;
import at.ac.uibk.gitsearch.service.PluginManagementService.PluginConfigWrapper.ActionWrapper;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.PluginActionInfo;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.SearchResultDTO;
import jakarta.el.ELContext;
import jakarta.el.ExpressionFactory;
import jakarta.el.StandardELContext;
import jakarta.el.ValueExpression;

/**
 * management services for plugins.
 * 
 * @author Michael Breu
 *
 */
@Service
public class PluginManagementService {

	private final class PluginCheckTimer extends TimerTask {
		private final String registeredPluginURL;

		private PluginCheckTimer(String registeredPlugin) {
			this.registeredPluginURL = registeredPlugin;
		}

		@Override
		public void run() {
			ClientConfig restClientConfig = new ClientConfig();
			restClientConfig.register(JacksonJsonProvider.class);
			Client client = ClientBuilder.newClient(restClientConfig);
			
			WebTarget target = client.target(registeredPluginURL);
			SharingPluginConfig config = null;
			try {
				config = target.request().accept(MediaType.APPLICATION_JSON)
						.get(SharingPluginConfig.class);
				registeredPluginConfigs.put(config.pluginName, new PluginConfigWrapper(config, registeredPluginURL) );
			} catch (Exception e) {
				log.error("Cannot (re-)load plugin at {}", registeredPluginURL, e);
			}
			
		}
	}

	/**
	 * this is just a wrapper, to provide a more efficient access to the original
	 * config.
	 * 
	 * @author Michael Breu
	 *
	 */
	public static class PluginConfigWrapper {

		
		public static class ActionWrapper {
			protected SharingPluginConfig.Action originalAction;
			protected ValueExpression filterExpression;
			protected PluginActionInfo pluginActionInfo;
			
			String absoluteActionTransferURL;
			

			public ActionWrapper(Action action, PluginConfigWrapper plugin) {
				super();
				this.originalAction = action;
				pluginActionInfo = new PluginActionInfo(plugin.getPluginName(), action.actionId, action.actionName);

				try {
					factory.createValueExpression(factory, Boolean.class);
					ELContext context = new StandardELContext(factory);
					filterExpression = factory.createValueExpression(context,
							"${" + originalAction.filterELExpression + "}", Boolean.class);
				} catch (Exception e) {
					log.error("Cannot parse elExpression {} for pluginAction {}:{} ", originalAction.filterELExpression,
							plugin.getPluginName(), action.actionId);
					filterExpression = null;
				}
				
				try {
					URI actionURI = new URI(action.actionTransferURL);
					if (!actionURI.isAbsolute()) {
						this.absoluteActionTransferURL = new URI(plugin.getBaseURL()).resolve(actionURI).toASCIIString();
					}
				} catch (URISyntaxException e) {
					log.error("Cannot asolutize url {} from action {} via {}", action.actionTransferURL, plugin.getBaseURL(), plugin.getPluginName(), e);
				}


			}

			
			/**
			 * @return the absoluteImportServiceURL
			 */
			public String getAbsoluteActionTransferURL() {
				return absoluteActionTransferURL;
			}


			/**
			 * @return the originalAction
			 */
			public SharingPluginConfig.Action getOriginalAction() {
				return originalAction;
			}

			public String getActionId() {
				return originalAction.actionId;
			}

			public String getActionName() {
				return originalAction.actionName;
			}

			public PluginActionInfo getPluginActionInfo() {
				return pluginActionInfo;
			}

			protected static ExpressionFactory factory = jakarta.el.ExpressionFactory.newInstance();

			public boolean isApplicable(final SearchResultDTO hit) {
				if (filterExpression == null)
					return false;
				try {
					StandardELContext context = new StandardELContext(factory);
					context.getELResolver().setValue(context, null, "metadata", hit);
					return (boolean) filterExpression.getValue(context);
				} catch (Exception e) {
					log.error("Cannot evaluate elExpression {} for pluginAction {}:{} ",
							originalAction.filterELExpression, originalAction.actionId, e);
					return false;
				}

			}
		}

		protected SharingPluginConfig originalConfig;
		protected Map<String, ActionWrapper> actions;
		protected String baseURL;

		public PluginConfigWrapper(SharingPluginConfig originalConfig, String baseURL) {
			super();
			this.originalConfig = originalConfig;
			this.baseURL = baseURL;
			actions = new HashMap<>();
			for (Action action : originalConfig.actions) {
				actions.put(action.actionId, new ActionWrapper(action, this));
			}
		}

		/**
		 * @return the baseURL
		 */
		public String getBaseURL() {
			return baseURL;
		}

		public String getPluginName() {
			return originalConfig.pluginName;
		}

		public Map<String, ActionWrapper> getActions() {
			return actions;
		}

	}
	
	@Autowired 
	private ApplicationProperties applicationProperties;

	private static Map<String, PluginConfigWrapper> registeredPluginConfigs = new HashMap<>();

	private static final Logger log = LoggerFactory.getLogger(PluginManagementService.class);

	private static final long PLUGIN_RECHECK_TIMEOUT = 5 * 60 * 1000L; // every 5 Minutes
	
	@PostConstruct
	public void registerPlugins() {
		 List<String> registeredPlugins = applicationProperties.getRegisteredPlugins();

		if (registeredPlugins==null) {
			log.warn("No Plugins registered?");
			return;
		}

		for (String registeredPlugin : registeredPlugins) {
			Timer check = new Timer(registeredPlugin+"-CheckTimer");
			check.scheduleAtFixedRate(new PluginCheckTimer(registeredPlugin),0L, PLUGIN_RECHECK_TIMEOUT);
			
		}
	}


	public String getRedirectURL(String pluginName, String actionId) {
		final PluginConfigWrapper sharingPluginConfig = registeredPluginConfigs.get(pluginName);
		if (sharingPluginConfig == null)
			return null;
		final ActionWrapper actionWrapper = sharingPluginConfig.actions.get(actionId);
		if (actionWrapper == null)
			return null;
		return actionWrapper.getAbsoluteActionTransferURL();
	}

	public Collection<PluginConfigWrapper> getRegisteredPluginConfigs() {
		return registeredPluginConfigs.values();
	}

}
